

#include "./stdint.h"
#include "HE_dos.h"

//like the micro does...
extern uint8_t secBuf[ 512 ];
//extern HEDskVolInfo* pVolInfo;

bool GetSectorData( uint32_t sector );
bool SetSectorData( uint32_t sector );

//not inlined in port, but this may not be the right thing to do
uint32_t HEDsk_GetCurFileSector( pHEDskFileInfo pInfo )
{
	return
		heVol.rootData
		+ ( pInfo->curClust - 2 ) * (uint16_t)heVol.secPerClust //this should be shift, not call to mult(..)
		+ ( pInfo->curSectorPos & ( heVol.secPerClust - 1 ) );
}

//basic succsess is 'GetSectorData for now...
//The MSP430 and x86 are little endian, so nothing special
bool HEDsk_LoadVolumeInfo( )
{
	//can't be zero to start cause of caching
	heVol.curSector= 0xffff;
	
	//Get Master Boot Record (MBR), sector 0
	if( ! GetSectorData( 0 ) )
		return false;

	//Get the sector of the first partition from the MBR
	heVol.bootSector= *(uint32_t*)( &secBuf[ MBR_LBA0 ] );

	//Get the FAT Boot Sector
	if( ! GetSectorData( heVol.bootSector ) )
		return false;

	//The root sector of the FAT table
	heVol.rootFat= heVol.bootSector + *(uint16_t*)( &secBuf[ HEFAT_RESERVED ] );

	//Root sector of the directory
	heVol.rootDir= heVol.rootFat + ( *(uint16_t*)( &secBuf[ HEFAT_SECPERFAT ] ) << 1 );

	//count of directory entries
	heVol.dirEntries= *(uint16_t*)( &secBuf[ HEFAT_DIRENTRIES ] );

	//Sectors per cluster
	heVol.secPerClust= secBuf[ HEFAT_SECPERCLST ];

	//Bytes per cluster for seek
	heVol.bytesPerClust= (uint16_t)heVol.secPerClust << HEDS_SECTOR_SHIFT;

	//Root sector of the data area
	heVol.rootData= heVol.rootDir + ( heVol.dirEntries >> 3 ) + 1;

	//for fileInfo get absolute sector
	//heVol.secPerClustShift= 0;
	//for( uint16_t temp= heVol.secPerClust; temp >>= 1; )
	//	++heVol.secPerClustShift;

	//Set the first free fat position to start
	// not really free, need to call HEDsk_GetFreeFat(...)
	//Fragmentation is not our concern in this app
	heVol.freeFatPos= 0x0002;

	//start with flags reset
	heVol.volFlags= 0;
	//and current sector
	heVol.curSector= 0;

	return true;
}

bool HEDsk_GetFileInfo( pHEDskFileInfo pInfo, uint8_t* psCanonicalName )
{
	//Get the root dir sector to start search for file
	GetSectorData( heVol.rootDir );

	uint32_t dirCount= 1;
	uint32_t dirSecCount= heVol.rootDir;
	uint8_t* pEnt= secBuf;
	for( ; dirCount < heVol.dirEntries; pEnt+= HEFAT_SZF_DIRENT )
	{
		if( ! ( ++dirCount & HEDS_DIRCNTMASK ) )
		{
			//need to move to next sector and reset pEnt;
			GetSectorData( ++dirSecCount );
			pEnt= secBuf;
		}
		if( ( pEnt[ 11 ] & 0x1f ) || *pEnt == 0xe5 )
			//We are just looking for real cananonical file names
			continue;

		if( HEDsk_strcmp( psCanonicalName, (uint8_t*)pEnt, HEFAT_SZF_FNAME ) )
		{
			//found it
			//ASSERT( pEnt - secBuf < HEDS_SECTORSIZE );
			pInfo->dirPos= (uint16_t)( pEnt - secBuf );
			pInfo->dirSect= (uint16_t)dirSecCount;
			pInfo->size= *(uint32_t*)( &pEnt[ HEDS_FILE_SIZE ] );
			//set file pointer to begining
			pInfo->curClust= *(uint16_t*)( &pEnt[ HEDS_FILE_DIR_CLUST ] );
			pInfo->curPointer= 0;
			pInfo->curSectorPos= 0;
			return true;
		}
	}
	return false;
}

void HEDsk_GetFreeFat( )
{
	//need sector and pos.
	uint32_t fatSec= (uint16_t)heVol.rootFat + ( heVol.freeFatPos >> HEDS_FATSECSHIFT );
	uint8_t fatPos= heVol.freeFatPos & HEDS_HALFSECMASK;

	for( ; ; )
	{
		//This caches so don't check refetch
		GetSectorData( fatSec );

		//Is cluster free?
		if( ! *(uint16_t*)&secBuf[ fatPos << 1 ] )
			break;

		if( ! ( ++fatPos & HEDS_ENDFATMASK ) )
		{
			//move to start of next fat sector
			fatPos= 0;
			++fatSec;
			//TODO ? check if partition is full ?!
		}
	}
	//save as a FAT entry
	heVol.freeFatPos= (uint16_t)( ( fatSec - heVol.rootFat ) << HEDS_FATSECSHIFT ) | fatPos;
	//mark it as an end cluster, only reason for finding it
	*(uint16_t*)&secBuf[ fatPos << 1 ]= HEDS_ENDCLUSTMSK;
	//Mark sector dirty
	SetSectorData( fatSec );
}

void HEDsk_AddCluster( pHEDskFileInfo pInfo )
{
	HEDsk_GetFreeFat( );

	//Is the file zero length?
	if( ! pInfo->curClust )
	{
		GetSectorData( pInfo->dirSect );
		//I've seen this non zero when it should be. Just me?
		//*(uint32_t*)&secBuf[ pInfo->dirPos + HEDS_FILE_SIZE ]= 0;
		*(uint16_t*)&secBuf[ pInfo->dirPos + HEDS_FILE_DIR_CLUST ]= heVol.freeFatPos;
		SetSectorData( pInfo->dirSect );
	}
	else
	{
		uint32_t curSector= ( pInfo->curClust >> HEDS_FATSECSHIFT )  + heVol.rootFat;
		GetSectorData( curSector );

		//update the last cluster entry in the FAT
		*(uint16_t*)&secBuf[ ( pInfo->curClust & HEDS_HALFSECMASK ) << 1 ]= heVol.freeFatPos;
		SetSectorData( curSector );
	}
	//Update the file info
	pInfo->curClust= heVol.freeFatPos;
}

//Warning: No check here that pos is <= size.
//No sector fetch, just sets the pHEDskFileInfo pointers
void HEDsk_SeekFile( pHEDskFileInfo pInfo, uint32_t pos )
{
	//This is the byte count offset into the current file position
	uint32_t absPos= pInfo->curPointer + ( (uint32_t)pInfo->curSectorPos << HEDS_SECTOR_SHIFT );

	if( absPos > pos )
	{
		//there is a start pos with cur cluster that pos may be greater than, but
		//rewind and start over
		absPos= 0;
		pInfo->curSectorPos= 0;
		GetSectorData( pInfo->dirSect );
		pInfo->curClust= *(uint16_t*)&secBuf[ pInfo->dirPos + HEDS_FILE_DIR_CLUST ];
	}

	//subPos is byte to go
	uint32_t subPos= pos - absPos;

	//first is to move by cluster
	for( ; subPos > heVol.bytesPerClust; )
	{
		uint32_t fsector= ( pInfo->curClust >> HEDS_FATSECSHIFT ) + heVol.rootFat;
		GetSectorData( fsector );
		pInfo->curClust= *(uint16_t*)( &secBuf[ ( pInfo->curClust & HEDS_SECTORMASK ) << 1 ] );

		pInfo->curSectorPos+= heVol.secPerClust;
		subPos-= heVol.bytesPerClust;
	}
	//At this point we should be in the cluster
	pInfo->curSectorPos+= subPos >> HEDS_SECTOR_SHIFT;
	pInfo->curPointer= (uint16_t)pos & HEDS_SECTORMASK;
}

bool HEDsk_ReadFile( pHEDskFileInfo pInfo, uint8_t* buf, uint16_t count )
{
	//Absolute sector is not kept, should it?
	uint32_t curSector= HEDsk_GetCurFileSector( pInfo );

	//out data pos
	uint8_t* pOutBuf= buf;

	for( ; count; )
	{
		uint16_t readSize;

		if( pInfo->curPointer + count > HEDS_SECTORSIZE )
			readSize= HEDS_SECTORSIZE - (uint16_t)pInfo->curPointer;
		else
			readSize= count;

		GetSectorData( curSector );
		HEDsk_memcpy( pOutBuf, secBuf + pInfo->curPointer, readSize );

		pInfo->curPointer+= readSize;
		if( pInfo->curPointer == HEDS_SECTORSIZE )
			pInfo->curPointer= 0;

		pOutBuf+= readSize;
		count-= readSize;

		//Need more sectors?
		if( ! count )
			break;

		++curSector
		//move into the next cluster?
		;
		if( ! ( ++pInfo->curSectorPos & ( heVol.secPerClust - 1 ) ) )
		{
			uint32_t fsector= ( pInfo->curClust >> HEDS_FATSECSHIFT ) + heVol.rootFat;
			GetSectorData( fsector );
			pInfo->curClust= *(uint16_t*)( &secBuf[ ( pInfo->curClust & HEDS_SECTORMASK ) << 1 ] );
			curSector= HEDsk_GetCurFileSector( pInfo );
		}
	}
	return true;
}

bool HEDsk_WriteFile( pHEDskFileInfo pInfo, uint8_t* pBuf, uint16_t count )
{
	//cause...
	if( ! count )
		return true;

	//working counter, need count later for dir
	uint16_t workCount= count;

	//Absolute sector is not kept, should it?
	uint32_t curSector= HEDsk_GetCurFileSector( pInfo );

	//where write starts
	uint32_t nextSize= pInfo->curPointer + count + ( pInfo->curSectorPos << HEDS_SECTOR_SHIFT );

	//special case, file length is zero. Don't trust 'size' ?
	if( ! pInfo->curClust )
		HEDsk_AddCluster( pInfo );

	//If we were at the end of the sector already, this didn't need to happen
	GetSectorData( curSector );

	for( ; workCount; )
	{
		uint16_t writeSize;
		if( pInfo->curPointer + workCount > HEDS_SECTORSIZE )
			writeSize= HEDS_SECTORSIZE - (uint16_t)pInfo->curPointer;
		else
			writeSize= workCount;
		
		HEDsk_memcpy( secBuf + pInfo->curPointer, pBuf, writeSize );
		pInfo->curPointer+= writeSize;
		SetSectorData( curSector );

		workCount-= writeSize;

		//Need more sectors?
		if( workCount )
		{
			//bump curSectorPos, move into the next cluster?
			if( ! ( ++pInfo->curSectorPos & ( heVol.secPerClust - 1 ) ) )
			{
				//Add the cluster
				HEDsk_AddCluster( pInfo );

				//Need the data sector back for subsequent writes
				curSector= HEDsk_GetCurFileSector( pInfo );
			}
			else
				//just move up through cluster
				++curSector;

			//freshen up for partial writes
			GetSectorData( curSector );
			pBuf+= writeSize;
			pInfo->curPointer= 0;
		}
	}
	//what if we are writing back into alloced part of file?
	if( nextSize > pInfo->size )
        pInfo->size= nextSize;

	return true;
}

bool HEDsk_FlushFile( pHEDskFileInfo pInfo )
{
	GetSectorData( pInfo->dirSect );
	*(uint32_t*)&secBuf[ pInfo->dirPos + HEDS_FILE_SIZE ]= pInfo->size;
	SetSectorData( pInfo->dirSect );
	GetSectorData( 0 );
	return true;
}

bool HEDsk_strcmp( const uint8_t* a, const uint8_t* b, uint16_t count )
{
	for( ; count--; )
	{
		if( *a != *b )
			return false;
		++a;
		++b;
	}
	return true;
}

void HEDsk_memcpy( uint8_t* pDest, uint8_t* pSource, uint16_t count )
{
	for( ; count--; )
	{
		*pDest++= *pSource++;
	}
}


/*	in port
uint32_t HEDsk_DskGetCurFileSector( pHEDskFileInfo pInfo )
{
	return
		heVol.rootData
		+ ( pInfo->curClust - 2 ) * (uint16_t)heVol.secPerClust //this should be shift, not call to mult(..)
		+ ( pInfo->curSectorPos & ( heVol.secPerClust - 1 ) );
}
*/

/*
#ifdef _WINDOWS //here we will set the date modified, doesn't work out
#include <SYS\UTIME.H>
#endif

bool HEDsk_FlushFile( pHEDskFileInfo pInfo )
{
//	GetSectorData( pInfo->dirSect );
//#ifdef _WINDOWS //here we will set the date modified, doesn't work out
//				//File time date
//				CTime dtDateMod;
//				dtDateMod= CTime::GetCurrentTime( );
//
//				__utimbuf64 tb;
//				tb.actime= dtDateMod.GetTime( );
//				tb.modtime= dtDateMod.GetTime( );
//
//				_utime64( "G:\\test.txt", &tb );
//#endif
	GetSectorData( pInfo->dirSect );
	*(uint32_t*)&secBuf[ pInfo->dirPos + HEDS_FILE_SIZE ]= pInfo->size;
	SetSectorData( pInfo->dirSect );
	GetSectorData( 0 );
	return true;
}
*/
